#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include "analyze.h"
#include "lib.h"

static int analyze_arp(FILE *fp, char *buffer, int size)
{
  struct arphdr arphdr;
  struct in_addr ipaddr;
  char *p = buffer;

  if (size < sizeof(arphdr))
    return -1;
  memcpy(&arphdr, buffer, sizeof(arphdr));

  fprintf(fp, "ARP");
  fprintf(fp, "\toperation: %d\n", ntohs(arphdr.ar_op));

  p += sizeof(arphdr);
  if (p - buffer + ETHER_ADDR_LEN > size) return -1;
  fprintf(fp, "\tsender MAC:%s\n", ether_ntoa((struct ether_addr *)p));

  p += arphdr.ar_hln;
  if (p - buffer + sizeof(ipaddr) > size) return -1;
  memcpy(&ipaddr, p, sizeof(ipaddr));
  fprintf(fp, "\tsender IP :%s\n", inet_ntoa(ipaddr));

  p += arphdr.ar_pln;
  if (p - buffer + ETHER_ADDR_LEN > size) return -1;
  fprintf(fp, "\ttarget MAC:%s\n", ether_ntoa((struct ether_addr *)p));

  p += arphdr.ar_hln;
  if (p - buffer + sizeof(ipaddr) > size) return -1;
  memcpy(&ipaddr, p, sizeof(ipaddr));
  fprintf(fp, "\ttarget IP :%s\n", inet_ntoa(ipaddr));

  return 0;
}

static int analyze_icmp(FILE *fp, char *buffer, int size, struct ip *iphdr)
{
  struct icmp *icmphdr;
  int s;

  if (size < sizeof(*icmphdr))
    return -1;
  icmphdr = (struct icmp *)buffer;
  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;

  fprintf(fp, "ICMP");
  fprintf(fp, "\ttotal size: %d bytes\n", s);
  fprintf(fp, "\ttype      : %d\n", icmphdr->icmp_type);
  fprintf(fp, "\tcode      : %d\n", icmphdr->icmp_code);
  fprintf(fp, "\tchecksum  : %04x\n", ntohs(icmphdr->icmp_cksum));

  return 0;
}

static int analyze_igmp(FILE *fp, char *buffer, int size, struct ip *iphdr)
{
  struct igmp *igmphdr;

  if (size < sizeof(*igmphdr))
    return -1;
  if (size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  fprintf(fp, "IGMP");
  fprintf(fp, "\ttype    : %d\n", igmphdr->igmp_type);
  fprintf(fp, "\tcode    : %d\n", igmphdr->igmp_code);
  fprintf(fp, "\tchecksum: %04x\n", ntohs(igmphdr->igmp_cksum));
  fprintf(fp, "\tgroup   : %s\n", inet_ntoa(igmphdr->igmp_group));

  return 0;
}

static int analyze_tcp(FILE *fp, char *buffer, int size, struct ip *iphdr)
{
  struct tcphdr *tcphdr;
  int s;

  if (size < sizeof(*tcphdr))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;

  fprintf(fp, "TCP");
  fprintf(fp, "\ttotal size: %d bytes\n", s);
  fprintf(fp, "\tsrc port  : %d\n", ntohs(tcphdr->th_sport));
  fprintf(fp, "\tdst port  : %d\n", ntohs(tcphdr->th_dport));
  fprintf(fp, "\tseq number: %u\n", ntohl(tcphdr->th_seq));
  fprintf(fp, "\tack number: %u\n", ntohl(tcphdr->th_ack));
  fprintf(fp, "\toffset    : %d\n", (tcphdr->th_off << 2));
  fprintf(fp, "\tflags     : %02x ( ", tcphdr->th_flags);
  if (tcphdr->th_flags & TH_FIN ) fprintf(fp, "FIN ");
  if (tcphdr->th_flags & TH_SYN ) fprintf(fp, "SYN ");
  if (tcphdr->th_flags & TH_RST ) fprintf(fp, "RST ");
  if (tcphdr->th_flags & TH_PUSH) fprintf(fp, "PSH ");
  if (tcphdr->th_flags & TH_ACK ) fprintf(fp, "ACK ");
  if (tcphdr->th_flags & TH_URG ) fprintf(fp, "URG ");
  fprintf(fp, ")\n");
  fprintf(fp, "\twindow    : %d\n", ntohs(tcphdr->th_win));
  fprintf(fp, "\tchecksum  : %04x\n", ntohs(tcphdr->th_sum));

  return 0;
}

static int analyze_udp(FILE *fp, char *buffer, int size, struct ip *iphdr)
{
  struct udphdr *udphdr;
  int s;

  if (size < sizeof(*udphdr))
    return -1;
  udphdr = (struct udphdr *)buffer;

  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;

  fprintf(fp, "UDP");
  fprintf(fp, "\ttotal size: %d bytes\n", s);
  fprintf(fp, "\tsrc port  : %d\n", ntohs(udphdr->uh_sport));
  fprintf(fp, "\tdst port  : %d\n", ntohs(udphdr->uh_dport));
  fprintf(fp, "\tlength    : %d\n", ntohs(udphdr->uh_ulen));
  fprintf(fp, "\tchecksum  : %04x\n", ntohs(udphdr->uh_sum));

  return 0;
}

static int analyze_ip(FILE *fp, char *buffer, int size)
{
  int r = 0;
  static char *pktbuf = NULL;
  static int bufsize = 0;
  struct ip *iphdr;
  int hdrsize;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;

  if (size < hdrsize)
    return -1;
  if (size < ntohs(iphdr->ip_len))
    return -1;

  fprintf(fp, "IP");
  fprintf(fp, "\theader size: %d bytes\n", hdrsize);
  fprintf(fp, "\ttotal size : %d bytes\n", ntohs(iphdr->ip_len));
  fprintf(fp, "\tID         : %04x\n", ntohs(iphdr->ip_id));
  fprintf(fp, "\tfragment   : %04x\n", ntohs(iphdr->ip_off));
  fprintf(fp, "\tTTL        : %d\n", iphdr->ip_ttl);
  fprintf(fp, "\tprotocol   : %d\n", iphdr->ip_p);
  fprintf(fp, "\tchecksum   : %04x\n", ntohs(iphdr->ip_sum));
  fprintf(fp, "\tsrc IP addr: %s\n", inet_ntoa(iphdr->ip_src));
  fprintf(fp, "\tdst IP addr: %s\n", inet_ntoa(iphdr->ip_dst));

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP:
    r = analyze_icmp(fp, pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_IGMP:
    r = analyze_igmp(fp, pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_TCP:
    r = analyze_tcp(fp, pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_UDP:
    r = analyze_udp(fp, pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  default:
    break;
  }

  return r;
}

int pkt_analyze_ethernet(FILE *fp, char *buffer, int size, struct timeval *tm)
{
  int r = 0;
  struct ether_header ehdr;
  union {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;
  static int count = 0;

  count++;
  fprintf(fp, "-- %d --\n", count);
  fprintf(fp, "received: %d bytes\n", size);
  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);
  fprintf(fp, "%s ->", ether_ntoa((struct ether_addr *)ehdr.ether_shost));
  fprintf(fp, " %s  ", ether_ntoa((struct ether_addr *)ehdr.ether_dhost));

  type = ntohs(ehdr.ether_type);
  buffer += ETHER_HDR_LEN;
  size   -= ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
  while (type == ETHERTYPE_VLAN) {
    if (size < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, buffer, sizeof(vlantag));
    fprintf(fp, "(VLAN tag: 0x%04x) ", ntohs(vlantag.tag));
    type = ntohs(vlantag.proto);
    buffer += sizeof(vlantag);
    size   -= sizeof(vlantag);
  }

  fprintf(fp, "(type: 0x%04x)\n", type);

  switch (type) {
  case ETHERTYPE_ARP:
    r = analyze_arp(fp, buffer, size);
    break;
  case ETHERTYPE_IP:
    r = analyze_ip(fp, buffer, size);
    break;
  default:
    break;
  }

  fprintf(fp, "==\n");
  fflush(fp);

  return r;
}
