#ifdef __linux__
#define _BSD_SOURCE
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>

#include <netinet/ip_icmp.h>
#include <netinet/igmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

#include "correct.h"
#include "lib.h"

static int calc_checksum(void *buffer, int size)
{
  union {
    char c[2];
    unsigned short s;
  } w;
  char *p;
  int sum = 0;

  for (p = buffer; size > 0; p += 2) {
    w.c[0] = p[0];
    w.c[1] = (size > 1) ? p[1] : 0;
    sum += ntohs(w.s);
    size -= 2;
  }
  sum = (sum & 0xffff) + (sum >> 16);
  sum = (sum & 0xffff) + (sum >> 16);

  return ((~sum) & 0xffff);
}

struct pseudo_header {
  long saddr;
  long daddr;
  char zero;
  char protocol;
  short len;
};

static int correct_icmp(char *buffer, int size, struct ip *iphdr)
{
  struct icmp *icmphdr;
  int s;

  if (size < sizeof(*icmphdr))
    return -1;
  icmphdr = (struct icmp *)buffer;
  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;

  /* This is compatible with FreeBSD */
  icmphdr->icmp_cksum = 0;
  icmphdr->icmp_cksum = htons(calc_checksum(icmphdr, s));

  return 0;
}

static int correct_igmp(char *buffer, int size, struct ip *iphdr)
{
  struct igmp *igmphdr;

  if (size < sizeof(*igmphdr))
    return -1;
  if (size < IGMP_MINLEN)
    return -1;
  igmphdr = (struct igmp *)buffer;

  /* This is compatible with FreeBSD */
  igmphdr->igmp_cksum = 0;
  igmphdr->igmp_cksum = htons(calc_checksum(igmphdr, sizeof(struct igmp)));

  return 0;
}

static int correct_tcp(char *buffer, int size, struct ip *iphdr)
{
  struct tcphdr *tcphdr;
  struct pseudo_header phdr;
  int s;

  if (size < sizeof(*tcphdr))
    return -1;
  tcphdr = (struct tcphdr *)buffer;

  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;
  phdr.saddr = iphdr->ip_src.s_addr;
  phdr.daddr = iphdr->ip_dst.s_addr;
  phdr.zero  = 0;
  phdr.protocol = iphdr->ip_p;
  phdr.len = iphdr->ip_len;

  /* This is compatible with FreeBSD */
  tcphdr->th_sum = htons(calc_checksum(&phdr, sizeof(phdr)));
  tcphdr->th_sum = htons(calc_checksum(tcphdr, s));

  return 0;
}

static int correct_udp(char *buffer, int size, struct ip *iphdr)
{
  struct udphdr *udphdr;
  struct pseudo_header phdr;
  int s;

  if (size < sizeof(*udphdr))
    return -1;
  udphdr = (struct udphdr *)buffer;

  s = ntohs(iphdr->ip_len) - (iphdr->ip_hl << 2);
  if (size < s)
    return -1;
  phdr.saddr = iphdr->ip_src.s_addr;
  phdr.daddr = iphdr->ip_dst.s_addr;
  phdr.zero  = 0;
  phdr.protocol = iphdr->ip_p;
  phdr.len = iphdr->ip_len;

  /* This is compatible with FreeBSD */
  udphdr->uh_sum = htons(calc_checksum(&phdr, sizeof(phdr)));
  udphdr->uh_sum = htons(calc_checksum(udphdr, s));

  return 0;
}

static int correct_ip(char *buffer, int size)
{
  int r = 0;
  static char *pktbuf = NULL;
  static int bufsize = 0;
  struct ip *iphdr;
  int hdrsize;

  pktbuf = pkt_alloc_buffer(pktbuf, &bufsize, size);

  if (size < sizeof(*iphdr))
    return -1;

  memcpy(pktbuf, buffer, size);
  iphdr = (struct ip *)pktbuf;

  hdrsize = iphdr->ip_hl << 2;

  if (size < hdrsize)
    return -1;
  if (size < ntohs(iphdr->ip_len))
    return -1;

  switch (iphdr->ip_p) {
  case IPPROTO_ICMP:
    r = correct_icmp(pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_IGMP:
    r = correct_igmp(pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_TCP:
    r = correct_tcp(pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  case IPPROTO_UDP:
    r = correct_udp(pktbuf + hdrsize, size - hdrsize, iphdr);
    break;
  default:
    break;
  }
  iphdr->ip_sum = 0;
  iphdr->ip_sum = htons(calc_checksum(iphdr, hdrsize));

  memcpy(buffer, pktbuf, size);

  return r;
}

int pkt_correct_ethernet(char *buffer, int size)
{
  int r = 0;
  struct ether_header ehdr;
  union {
    pkt_uint16 tag;
    pkt_uint16 proto;
  } vlantag;
  int type;

  if (size < ETHER_HDR_LEN)
    return -1;
  memcpy(&ehdr, buffer, ETHER_HDR_LEN);

  type = ntohs(ehdr.ether_type);
  buffer += ETHER_HDR_LEN;
  size   -= ETHER_HDR_LEN;

#ifndef ETHERTYPE_VLAN
#define ETHERTYPE_VLAN 0x8100
#endif
  while (type == ETHERTYPE_VLAN) {
    if (size < sizeof(vlantag))
      return -1;
    memcpy(&vlantag, buffer, sizeof(vlantag));
    type = ntohs(vlantag.proto);
    buffer += sizeof(vlantag);
    size   -= sizeof(vlantag);
  }

  switch (type) {
  case ETHERTYPE_IP:
    r = correct_ip(buffer, size);
    break;
  default:
    break;
  }

  return r;
}
